"""HDF5 package of the atlas-ftag-tools."""

from __future__ import annotations

from .h5add_col import h5_add_column
from .h5reader import H5Reader
from .h5utils import cast_dtype, get_dtype, join_structured_arrays, structured_from_dict
from .h5writer import H5Writer

__all__ = [
    "H5Reader",
    "H5Writer",
    "cast_dtype",
    "get_dtype",
    "h5_add_column",
    "join_structured_arrays",
    "structured_from_dict",
]
