"""Utils package of the atlas-ftag-tools."""

from __future__ import annotations

from .logging import logger, set_log_level
from .metrics import (
    calculate_efficiency,
    calculate_efficiency_error,
    calculate_rejection,
    calculate_rejection_error,
    get_discriminant,
    save_divide,
    weighted_percentile,
)

__all__ = [
    "calculate_efficiency",
    "calculate_efficiency_error",
    "calculate_rejection",
    "calculate_rejection_error",
    "get_discriminant",
    "logger",
    "save_divide",
    "set_log_level",
    "weighted_percentile",
]
